#ifndef __VMML__BLAS_TYPES__HPP__
#define __VMML__BLAS_TYPES__HPP__


namespace vmml
{
	
	namespace blas
	{
		
#ifdef __APPLE__
		
		typedef int  blas_int;
		
#else
		
		typedef int  blas_int;
		
#endif
		
		
	} // namespace blas
	
	
} // namespace vmml

#endif

