#ifndef __VMML__VISIBILITY__HPP__
#define __VMML__VISIBILITY__HPP__

namespace vmml
{

enum Visibility
{
    VISIBILITY_NONE     = 0,
    VISIBILITY_PARTIAL  = 1,
    VISIBILITY_FULL     = 2
};

} // namespace vmml

#endif

