/* 
* VMMLib - Vector & Matrix Math Lib
*  
* @author Jonas Boesch
* @author Stefan Eilemann
* @author Renato Pajarola
* @author Susanne Suter
* @author Philipp Schlegel
* @author Daniel Pfeifer
* @author David H. Eberly ( Wild Magic )
* @author Andrew Willmott ( VL )
*
* @license revised BSD license, check LICENSE
*
* parts of the source code of VMMLib were inspired by David Eberly's 
* Wild Magic and Andrew Willmott's VL.
* 
*/ 


#ifndef __VMML__VMMLIB__HPP__
#define __VMML__VMMLIB__HPP__

#include <vmmlib/vector.hpp>
#include <vmmlib/matrix.hpp>
#include <vmmlib/quaternion.hpp>
#include <vmmlib/frustum.hpp>
#include <vmmlib/frustum_culler.hpp>

#endif

